/*
  
  test_k_ascent_permutations: Unit tests for the code in kascentpermutation.c.

*/

#ifdef HAVE_CONFIG_H
  #include"config.h"
#endif

#ifdef HAVE_STDIO_H
   #include<stdio.h>
#endif

#ifdef HAVE_MATH_H
   #include<math.h>
#endif

#ifdef HAVE_ASSERT_H
   #include<assert.h>
#endif

#ifdef HAVE_STRING_H
   #include<string.h>
#endif

#ifdef HAVE_STDINT_H
   #include<stdint.h>
#endif

#ifdef HAVE_STDLIB_H
   #include<stdlib.h>
#endif

#ifdef HAVE_STDBOOL_H
   #include<stdbool.h>
#endif

#ifdef HAVE_TIME_H
   #include<time.h>
#endif

#ifdef HAVE_GSL_GSL_RNG_H
#include <gsl/gsl_rng.h>
#endif

#ifdef HAVE_GSL_GSL_RANDIST_H
#include <gsl/gsl_randist.h>
#endif

#include<argtable2.h>


/* Most of the functions in kascentpermutation.c aren't exposed, so we 
   include prototypes here. */

double new_ascent_probability(int n,int k);

typedef struct choicestruct {
  int  n;
  int  k;
} perm_choice_t;

void   build_choice_vector(int N, int K,
   			   int *nchoices,perm_choice_t **cv,
			   gsl_rng *r);

void random_k_ascent_permutation(int n,int k,
				 int *perm,
				 gsl_rng *rng);

int PD_VERBOSE=50;

bool nap_lookup_table_test()
{ 

  printf("----------------------------------------------\n"
	 "testing lookup table for new_ascent_prob values \n"
	 "----------------------------------------------\n");

  printf("testing the edge case nap(2,1) == 1.0...");

  if (fabs(new_ascent_probability(2,1)-1.0) > 1e-12) {

    printf("FAIL (%g != %g)\n",new_ascent_probability(2,1),1.0);
    return false;

  }

  printf("pass\n\n");

  printf("testing the line nap(3,1) ... nap(3,2) ...\n");
  
  {
    double expected_values3[2] =
      {0.5,1.};
  
    int i;
    for(i=1;i<=2;i++) {
    
      printf("\t nap(3,%d) = %g ... ",i,new_ascent_probability(3,i));

      if (fabs(new_ascent_probability(3,i) - expected_values3[i-1]) > 1e-12) {
      
	printf("FAIL (!= %g)\n",expected_values3[i-1]);
	return false;
      
      } else {

	printf("pass\n");

      }

    }
    
  }

  printf("testing the line nap(3,1) ... nap(3,2) ... pass\n\n");

  printf("testing the line nap(4,1) ... nap(4,3) ...\n");
  
  {
    double expected_values4[3] =
      {0.2727272727272727,0.7272727272727273,1.};
  
    int i;
    for(i=1;i<=3;i++) {
    
      printf("\t nap(4,%d) = %g ... ",i,new_ascent_probability(4,i));

      if (fabs(new_ascent_probability(4,i) - expected_values4[i-1]) > 1e-12) {
      
	printf("FAIL (!= %g)\n",expected_values4[i-1]);
	return false;
      
      } else {

	printf("pass\n");

      }

    }
    
  }

  printf("testing the line nap(4,1) ... nap(4,3) ... pass\n\n");



  printf("testing the line nap(9,1) ... nap(9,8) ...\n");
  
  double expected_values[8] =
    {0.01593625498007968,0.11835980284775466,
     0.291928281614797,0.5,0.708071718385203,
     0.8816401971522454,0.9840637450199203,1.};
  
  int i;
  for(i=1;i<=8;i++) {
    
    printf("\t nap(9,%d) = %g ... ",i,new_ascent_probability(9,i));

    if (fabs(new_ascent_probability(9,i) - expected_values[i-1]) > 1e-12) {
      
      printf("FAIL (!= %g)\n",expected_values[i-1]);
      return false;
      
    } else {

      printf("pass\n");

    }

  }

  printf("testing the line nap(9,1) ... nap(9,8) ... pass\n\n");

  printf("testing the line nap(10,1) ... nap(10,9) ...\n");
  
  double expected_values10[9] =
    {0.008884501480750246,0.08394648829431438,
     0.22464366684827503,0.4040160139931652,
     0.5959839860068348,0.775356333151725,
     0.9160535117056856,0.9911154985192497,1.};
  
  for(i=1;i<=9;i++) {
    
    printf("\t nap(10,%d) = %g ... ",i,new_ascent_probability(10,i));

    if (fabs(new_ascent_probability(10,i) - expected_values10[i-1]) > 1e-12) {
      
      printf("FAIL (!= %g)\n",expected_values10[i-1]);
      return false;
      
    } else {

      printf("pass\n");

    }

  }

  printf("testing the line nap(10,1) ... nap(10,9) ... pass\n\n");

  printf("testing the edge case nap(1000,1) ... nap(1000,999) ...");

  double expected_values1000[999] =
  {9.323303548847157e-299,4.0442825536072555e-174,
   3.826813481681508e-123,3.0632774861811816e-95,
   1.3110130924059691e-77,1.8726020884380376e-65,
   1.4451210622137992e-56,8.72768850666952e-50,
   1.9246004560634974e-44,4.008206023665234e-40,
   1.4629945795105282e-36,1.4238670476646839e-33,
   4.962384697910936e-31,7.665235491882494e-29,
   6.1464077983523e-27,2.8836091401051325e-25,
   8.680022909390123e-24,1.8019516473401666e-22,
   2.7322516397567848e-21,3.168815911326143e-20,
   2.918618350899925e-19,2.2017143782454887e-18,
   1.39560023430172e-17,7.594017883125466e-17,
   3.6118134276751584e-16,1.5246474113952038e-15,
   5.7873785841841514e-15,1.9977524725362876e-14,
   6.332342713087733e-14,1.8587088421145921e-13,
   5.089529050175372e-13,1.3084701543584258e-12,
   3.176401334911569e-12,7.31763766421388e-12,
   1.606951806231289e-11,3.377156866083491e-11,
   6.816352067252843e-11,1.3255040738291856e-10,
   2.490430814132113e-10,4.532583919147246e-10,
   8.00942020029598e-10,1.377054243918182e-9,
   2.3079297948366644e-9,3.777182388927345e-9,
   6.046067594090683e-9,9.47904011298139e-9,
   1.4575239926363044e-8,2.2006613782707137e-8,
   3.266332326779431e-8,4.770700850961898e-8,
   6.863236913072584e-8,9.733739418093338e-8,
   1.3620242427343695e-7,1.8817806655456322e-7,
   2.5688181149679733e-7,3.46703025078488e-7,
   4.6291581594552703e-7,6.117991096447898e-7,
   8.007631055556581e-7,1.0384811507328412e-6,
   1.335025941843791e-6,1.7020088643310897e-6,
   2.152721198683354e-6,2.7022758683295603e-6,
   3.367748372532026e-6,4.168315540429888e-6,
   5.1253907581759565e-6,6.262754358509316e-6,
   7.606677919294982e-6,9.186041292400947e-6,
   0.000011032441274402781,0.000013180290933493448,
   0.00001566690872005279,0.000018532596609023904,
   0.000021820706648045266,0.000025577695413808114,
   0.00002985316600811692,0.00003469989735258574,
   0.000040173860664977604,0.00004633422311928825,
   0.00005324333880442494,0.00006096672720160917,
   0.00006957303949754184,0.00007913401313823825,
   0.00008972441510680601,0.0001014219744770316,
   0.00011430730485337171,0.00012846381735687927,
   0.0001439776248559382,0.00016093743817076063,
   0.00017943445500183982,0.00019956224234545967,
   0.00022141661316450303,0.00024509549808079815,
   0.00027069881284673875,0.00029832832233957397,
   0.000328087501802268,0.0003600813960308287,
   0.0003944164771801708,0.0004312005018295163,
   0.0004705423679146739,0.0005125519720988198,
   0.0005573400681161752,0.0006050181265847223,
   0.0006556981967452701,0.0007094927705451818,
   0.0007665146494462741,0.0008268768142981035,
   0.000890692298580371,0.0009580740652817164,
   0.0010291348876469617,0.0011039872339910666,
   0.0011827431567457967,0.0012655141858745067,
   0.0013524112267615654,0.0014435444626558517,
   0.0015390232617224725,0.0016389560887333956,
   0.0017434504214060276,0.001852612671378924,
   0.001966548109795692,0.0020853607974517605,
   0.0022091535194439225,0.002338027724249376,
   0.002472083467149353,0.0026114193579021643,
   0.0027561325125616616,0.0029063185093294954,
   0.003062071348323198,0.0032234834151368143,
   0.0033906454480666084,0.0035636465088710653,
   0.0037425739569320444,0.003927513426682338,
   0.004118548808164054,0.004315762230582025,
   0.004519234048716904,0.004729042832063539,
   0.004945265356561615,0.005167976598787479,
   0.0053972497324782,0.005633156127261541,
   0.005875765349468262,0.006125145164906305,
   0.006381361543479586,0.006644478665537604,
   0.006914558929845549,0.007191662963068284,
   0.00747584963066524,0.007767176049097054,
   0.008065697599248545,0.008371467940976386,
   0.00868453902869368,0.009004961127907298,
   0.009332782832627633,0.009668051083574002,
   0.0100108111871026,0.010361106834787371,
   0.010718980123587607,0.011084471576539492,
   0.011457620163912023,0.011838463324770912,
   0.01222703698889722,0.01262337559901032,
   0.013027512133247791,0.013439478127857524,
   0.013859303700060022,0.014287017571041442,
   0.014722647089040372,0.015166218252493695,
   0.01561775573320918,0.01607728289953454,
   0.016544821839494846,0.01702039338387205,
   0.017504017129202326,0.017995711460668713,
   0.01849549357486821,0.019003379502434043,
   0.01951938413049557,0.020043521224959376,
   0.02057580345259688,0.02111624240292476,
   0.021664848609865926,0.022221631573179853,
   0.022786599779652174,0.02335976072403451,
   0.023941120929726452,0.024530685969192525,
   0.025128460484107856,0.025734448205227013,
   0.026348651971971313,0.026971073751730558,
   0.027601714658875794,0.02824057497348044,
   0.028887654159747464,0.029542950884141158,
   0.030206463033222258,0.030878187731185735,
   0.03155812135710119,0.03224625956185579,
   0.03294259728480053,0.03364712877010054,
   0.034359847582790724,0.03508074662453825,
   0.035809818149113555,0.036547053777572065,
   0.037292444513148656,0.0380459807558675,
   0.03880765231686978,0.03957744843246213,
   0.04035535777788884,0.041141368480830734,
   0.04193546813463414,0.04273764381127305,
   0.043547882074048166,0.04436616899002604,
   0.04519249014222213,0.04602683064153128,
   0.04686917513840931,0.047719507834309526,
   0.04857781249287789,0.04944407245091064,
   0.050318270629078156,0.05120038954241899,
   0.052090411310607806,0.052988317668001154,
   0.05389408997346486,0.05480770921998688,
   0.05572915604407946,0.056658410734974414,
   0.057595453243615265,0.058540263191449914,
   0.059492819879027874,0.060453102294405356,
   0.06142108912136228,0.06239675874743442,
   0.06338008927176472,0.06437105851277695,
   0.06536964401567544,0.06637582305977434,
   0.06738957266565967,0.0684108696021878,
   0.06943969039332336,0.07047601132482033,
   0.07151980845074896,0.07257105759987226,
   0.07362973438187484,0.07469581419344729,
   0.07576927222422929,0.07685008346261418,
   0.07793822270141809,0.0790336645434165,
   0.08013638340675101,0.08124635353020929,
   0.08236354897838055,0.08348794364668985,
   0.08461951126631309,0.08575822540897604,
   0.08690405949163943,0.08805698678107282,
   0.08921698039831967,0.09038401332305596,
   0.0915580583978446,0.0927390883322884,
   0.0939270757070831,0.09512199297797341,
   0.0963238124796138,0.09753250642933614,
   0.09874804693082667,0.09997040597771388,
   0.10119955545706945,0.10243546715282432,
   0.10367811274910176,0.10492746383346904,
   0.10618349190011007,0.10744616835292015,
   0.10871546450852514,0.1099913515992264,
   0.11127380077587357,0.11256278311066614,
   0.11385826959988628,0.11516023116656383,
   0.11646863866307533,0.11778346287367845,
   0.11910467451698338,0.12043224424836244,
   0.12176614266229967,0.12310634029468129,
   0.12445280762502886,0.1258055150786761,
   0.12716443302889066,0.12852953179894255,
   0.12990078166411978,0.13127815285369296,
   0.13266161555282957,0.13405113990445974,
   0.13544669601109366,0.13684825393659267,
   0.13825578370789463,0.13966925531669433,
   0.14108863872108107,0.14251390384713272,
   0.14394502059046882,0.1453819588177628,
   0.14682468836821458,0.148273179054984,
   0.14972740066658688,0.15118732296825338,
   0.1526529157032508,0.15412414859417037,
   0.15560099134417998,0.1570834136382427,
   0.1585713851443027,0.1600648755144384,
   0.16156385438598478,0.16306829138262413,
   0.16457815611544743,0.16609341818398574,
   0.16761404717721315,0.16914001267452175,
   0.1706712842466688,0.17220783145669755,
   0.1737496238608314,0.17529663100934292,
   0.1768488224473975,0.17840616771587275,
   0.17996863635215388,0.18153619789090605,
   0.18310882186482347,0.18468647780535655,
   0.1862691352434171,0.18785676371006257,
   0.18944933273715886,0.19104681185802364,
   0.19264917060804893,0.19425637852530508,
   0.19586840515112536,0.197485220030672,
   0.1991067927134843,0.20073309275400897,
   0.20236408971211306,0.2039997531535803,
   0.20564005265059068,0.2072849577821839,
   0.20893443813470708,0.21058846330224712,
   0.212247002887048,0.21391002649991325,
   0.21557750376059423,0.21724940429816392,
   0.21892569775137774,0.22060635376902002,
   0.22229134201023815,0.22398063214486336,
   0.22567419385371934,0.22737199682891837,
   0.2290740107741456,0.23078020540493144,
   0.23249055044891287,0.23420501564608334,
   0.2359235707490317,0.23764618552317057,
   0.23937282974695454,0.2411034732120873,
   0.24283808572372004,0.24457663710063907,
   0.24631909717544453,0.24806543579471965,
   0.24981562281919056,0.25156962812387795,
   0.25332742159823946,0.2550889731463037,
   0.2568542526867961,0.25862323015325683,
   0.26039587549415005,0.2621721586729666,
   0.26395204966831787,0.26573551847402344,
   0.2675225350991903,0.2693130695682863,
   0.2711070919212059,0.27290457221332876,
   0.27470548051557314,0.27650978691444145,
   0.2783174615120603,0.2801284744262144,
   0.28194279579037373,0.2837603957537158,
   0.28558124448114147,0.28740531215328563,
   0.289232568966522,0.29106298513296275,
   0.2928965308804531,0.29473317645256025,
   0.29657289210855825,0.2984156481234065,
   0.30026141478772533,0.30211016240776545,
   0.3039618613053734,0.3058164818179526,
   0.3076739942984199,0.3095343691151582,
   0.31139757665196416,0.31326358730799303,
   0.31513237149769807,0.31700389965076775,
   0.3188781422120575,0.32075506964151923,
   0.32263465241412603,0.3245168610197943,
   0.3264016659633019,0.32828903776420293,
   0.33017894695673944,0.3320713640897502,
   0.33396625972657557,0.33586360444495983,
   0.3377633688369505,0.3396655235087947,
   0.3415700390808326,0.34347688618738814,
   0.34538603547665697,0.3472974576105918,
   0.34921112326478504,0.3511270031283495,
   0.35304506790379553,0.35496528830690655,
   0.35688763506661236,0.3588120789248595,
   0.36073859063647973,0.3626671409690565,
   0.3645977007027891,0.3665302406303545,
   0.3684647315567674,0.3704011442992387,
   0.37233944968703125,0.37427961856131425,
   0.3762216217750155,0.3781654301926724,
   0.38011101469028047,0.3820583461551405,
   0.38400739548570456,0.3859581335914195,
   0.3879105313925695,0.3898645598201164,
   0.3918201898155401,0.393777392330675,
   0.39573613832754767,0.39769639877821117,
   0.3996581446645792,0.40162134697825846,
   0.40358597672037944,0.4055520049014268,
   0.40751940254106794,0.40948814066798045,
   0.41145819031967884,0.4134295225423391,
   0.41540210839062436,0.4173759189275064,
   0.4193509252240892,0.4213270983594293,
   0.42330440942035685,0.42528282950129437,
   0.4272623297040753,0.42924288113776177,
   0.4312244549184615,0.43320702216914336,
   0.43519055401945295,0.4371750216055268,
   0.43916039606980606,0.4411466485608498,
   0.44313375023314694,0.4451216722469281,
   0.44711038576797685,0.44909986196743956,
   0.45109007202163603,0.4530809871118683,
   0.4550725784242295,0.45706481714941205,
   0.45905767448251583,0.46105112162285494,
   0.46304512977376516,0.46503967014240966,
   0.4670347139395858,0.46903023237953045,
   0.4710261966797254,0.47302257806070214,
   0.475019347745847,0.47701647696120475,
   0.47901393693528405,0.48101169889885964,
   0.48300973408477743,0.4850080137277569,
   0.48700650906419524,0.4890051913319697,
   0.4910040317702407,0.4930030016192547,
   0.49500207212014724,0.49700121451474455,
   0.49900040004536744,0.5009995999546325,
   0.5029987854852553,0.5049979278798529,
   0.5069969983807453,0.5089959682297593,
   0.5109948086680303,0.5129934909358046,
   0.5149919862722431,0.5169902659152226,
   0.5189883011011404,0.5209860630647161,
   0.5229835230387951,0.5249806522541531,
   0.5269774219392978,0.5289738033202745,
   0.5309697676204694,0.5329652860604142,
   0.5349603298575903,0.5369548702262349,
   0.5389488783771449,0.5409423255174841,
   0.542935182850588,0.5449274215757706,
   0.5469190128881316,0.5489099279783639,
   0.5509001380325604,0.5528896142320232,
   0.5548783277530718,0.5568662497668532,
   0.5588533514391502,0.5608396039301938,
   0.5628249783944733,0.5648094459805469,
   0.5667929778308566,0.5687755450815385,
   0.5707571188622382,0.5727376702959248,
   0.5747171704987056,0.576695590579643,
   0.5786729016405706,0.5806490747759109,
   0.5826240810724936,0.5845978916093757,
   0.5865704774576608,0.5885418096803212,
   0.5905118593320196,0.5924805974589321,
   0.5944479950985733,0.5964140232796207,
   0.5983786530217416,0.6003418553354208,
   0.6023036012217888,0.6042638616724524,
   0.606222607669325,0.60817981018446,
   0.6101354401798835,0.6120894686074305,
   0.6140418664085804,0.6159926045142955,
   0.6179416538448596,0.6198889853097196,
   0.6218345698073275,0.6237783782249844,
   0.6257203814386857,0.6276605503129687,
   0.6295988557007612,0.6315352684432326,
   0.6334697593696457,0.6354022992972108,
   0.6373328590309434,0.6392614093635203,
   0.6411879210751404,0.6431123649333876,
   0.6450347116930935,0.6469549320962045,
   0.6488729968716505,0.6507888767352149,
   0.6527025423894083,0.654613964523343,
   0.6565231138126117,0.6584299609191673,
   0.6603344764912052,0.6622366311630495,
   0.6641363955550401,0.6660337402734243,
   0.6679286359102498,0.6698210530432606,
   0.6717109622357972,0.673598334036698,
   0.6754831389802057,0.677365347585874,
   0.6792449303584808,0.6811218577879424,
   0.6829961003492322,0.6848676285023019,
   0.6867364126920071,0.6886024233480359,
   0.6904656308848419,0.6923260057015801,
   0.6941835181820475,0.6960381386946266,
   0.6978898375922346,0.6997385852122746,
   0.7015843518765935,0.7034271078914418,
   0.7052668235474397,0.7071034691195469,
   0.7089370148670373,0.710767431033478,
   0.7125946878467144,0.7144187555188585,
   0.7162396042462843,0.7180572042096264,
   0.7198715255737856,0.7216825384879396,
   0.7234902130855586,0.7252945194844269,
   0.7270954277866711,0.7288929080787941,
   0.7306869304317137,0.7324774649008097,
   0.7342644815259766,0.736047950331682,
   0.7378278413270335,0.7396041245058499,
   0.7413767698467432,0.7431457473132037,
   0.7449110268536963,0.7466725784017605,
   0.748430371876122,0.7501843771808094,
   0.7519345642052805,0.7536809028245555,
   0.755423362899361,0.75716191427628,
   0.7588965267879128,0.7606271702530455,
   0.7623538144768294,0.7640764292509683,
   0.7657949843539167,0.7675094495510871,
   0.7692197945950686,0.7709259892258543,
   0.7726280031710816,0.7743258061462807,
   0.7760193678551367,0.7777086579897619,
   0.7793936462309801,0.7810743022486223,
   0.782750595701836,0.7844224962394057,
   0.7860899735000866,0.787752997112952,
   0.7894115366977529,0.791065561865293,
   0.7927150422178161,0.7943599473494094,
   0.7960002468464197,0.7976359102878869,
   0.799266907245991,0.8008932072865157,
   0.802514779969328,0.8041315948488746,
   0.8057436214746949,0.8073508293919511,
   0.8089531881419765,0.8105506672628411,
   0.8121432362899376,0.8137308647565829,
   0.8153135221946435,0.8168911781351764,
   0.8184638021090939,0.8200313636478461,
   0.8215938322841272,0.8231511775526025,
   0.8247033689906571,0.8262503761391686,
   0.8277921685433025,0.8293287157533311,
   0.8308599873254783,0.8323859528227868,
   0.8339065818160142,0.8354218438845525,
   0.8369317086173758,0.8384361456140151,
   0.8399351244855616,0.8414286148556973,
   0.8429165863617574,0.84439900865582,
   0.8458758514058297,0.8473470842967492,
   0.8488126770317467,0.8502725993334131,
   0.851726820945016,0.8531753116317854,
   0.8546180411822372,0.8560549794095311,
   0.8574860961528673,0.8589113612789189,
   0.8603307446833056,0.8617442162921054,
   0.8631517460634073,0.8645533039889064,
   0.8659488600955403,0.8673383844471704,
   0.868721847146307,0.8700992183358803,
   0.8714704682010574,0.8728355669711093,
   0.8741944849213238,0.875547192374971,
   0.8768936597053187,0.8782338573377003,
   0.8795677557516376,0.8808953254830166,
   0.8822165371263216,0.8835313613369248,
   0.8848397688334362,0.8861417304001138,
   0.887437216889334,0.8887261992241265,
   0.8900086484007735,0.8912845354914749,
   0.8925538316470798,0.8938165080998899,
   0.8950725361665309,0.8963218872508983,
   0.8975645328471757,0.8988004445429305,
   0.900029594022286,0.9012519530691733,
   0.9024674935706639,0.9036761875203863,
   0.9048780070220266,0.9060729242929167,
   0.9072609116677116,0.9084419416021554,
   0.909615986676944,0.9107830196016802,
   0.911943013218927,0.9130959405083605,
   0.9142417745910241,0.9153804887336868,
   0.9165120563533101,0.9176364510216194,
   0.9187536464697907,0.919863616593249,
   0.9209663354565836,0.9220617772985819,
   0.9231499165373858,0.9242307277757708,
   0.9253041858065527,0.9263702656181252,
   0.9274289424001277,0.928480191549251,
   0.9295239886751797,0.9305603096066766,
   0.9315891303978123,0.9326104273343403,
   0.9336241769402258,0.9346303559843246,
   0.935628941487223,0.9366199107282351,
   0.9376032412525656,0.9385789108786378,
   0.9395468977055946,0.9405071801209721,
   0.9414597368085501,0.9424045467563849,
   0.9433415892650256,0.9442708439559206,
   0.9451922907800131,0.9461059100265352,
   0.9470116823319987,0.9479095886893921,
   0.9487996104575809,0.9496817293709219,
   0.9505559275490895,0.9514221875071222,
   0.9522804921656904,0.9531308248615907,
   0.9539731693584687,0.954807509857778,
   0.955633831009974,0.9564521179259519,
   0.9572623561887269,0.9580645318653659,
   0.9588586315191692,0.959644642222111,
   0.9604225515675379,0.9611923476831302,
   0.9619540192441325,0.9627075554868513,
   0.9634529462224279,0.9641901818508865,
   0.9649192533754617,0.9656401524172092,
   0.9663528712298995,0.9670574027151995,
   0.9677537404381443,0.9684418786428989,
   0.9691218122688142,0.9697935369667777,
   0.9704570491158588,0.9711123458402526,
   0.9717594250265196,0.9723982853411243,
   0.9730289262482694,0.9736513480280288,
   0.974265551794773,0.9748715395158922,
   0.9754693140308075,0.9760588790702737,
   0.9766402392759654,0.9772134002203477,
   0.9777783684268202,0.9783351513901342,
   0.9788837575970752,0.9794241965474031,
   0.9799564787750407,0.9804806158695044,
   0.980996620497566,0.9815045064251318,
   0.9820042885393312,0.9824959828707976,
   0.9829796066161279,0.9834551781605051,
   0.9839227171004654,0.9843822442667907,
   0.9848337817475062,0.9852773529109595,
   0.9857129824289587,0.98614069629994,
   0.9865605218721426,0.9869724878667522,
   0.9873766244009897,0.9877729630111028,
   0.9881615366752291,0.988542379836088,
   0.9889155284234605,0.9892810198764124,
   0.9896388931652127,0.9899891888128973,
   0.9903319489164261,0.9906672171673723,
   0.9909950388720928,0.9913154609713063,
   0.9916285320590236,0.9919343024007515,
   0.9922328239509028,0.9925241503693348,
   0.9928083370369317,0.9930854410701545,
   0.9933555213344624,0.9936186384565204,
   0.9938748548350937,0.9941242346505317,
   0.9943668438727384,0.9946027502675218,
   0.9948320234012126,0.9950547346434384,
   0.9952709571679365,0.9954807659512832,
   0.9956842377694181,0.9958814511918359,
   0.9960724865733177,0.9962574260430679,
   0.996436353491129,0.9966093545519333,
   0.9967765165848632,0.9969379286516769,
   0.9970936814906705,0.9972438674874384,
   0.9973885806420978,0.9975279165328507,
   0.9976619722757507,0.997790846480556,
   0.9979146392025483,0.9980334518902042,
   0.9981473873286211,0.9982565495785939,
   0.9983610439112667,0.9984609767382775,
   0.9985564555373442,0.9986475887732384,
   0.9987344858141255,0.9988172568432542,
   0.9988960127660089,0.9989708651123529,
   0.9990419259347183,0.9991093077014197,
   0.9991731231857018,0.9992334853505537,
   0.9992905072294548,0.9993443018032548,
   0.9993949818734154,0.9994426599318839,
   0.9994874480279011,0.9995294576320853,
   0.9995687994981706,0.9996055835228199,
   0.9996399186039692,0.9996719124981978,
   0.9997016716776603,0.9997293011871533,
   0.9997549045019192,0.9997785833868356,
   0.9998004377576544,0.9998205655449981,
   0.9998390625618293,0.9998560223751439,
   0.9998715361826432,0.9998856926951467,
   0.999898578025523,0.9999102755848931,
   0.9999208659868618,0.9999304269605025,
   0.9999390332727985,0.9999467566611957,
   0.9999536657768806,0.9999598261393351,
   0.9999653001026475,0.9999701468339919,
   0.9999744223045861,0.999978179293352,
   0.9999814674033909,0.99998433309128,
   0.9999868197090664,0.9999889675587256,
   0.9999908139587075,0.9999923933220808,
   0.9999937372456414,0.9999948746092419,
   0.9999958316844597,0.9999966322516274,
   0.9999972977241315,0.9999978472788014,
   0.9999982979911357,0.9999986649740581,
   0.9999989615188493,0.9999991992368944,
   0.9999993882008904,0.999999537084184,
   0.9999996532969748,0.9999997431181885,
   0.9999998118219334,0.9999998637975758,
   0.999999902662606,0.9999999313676308,
   0.9999999522929914,0.9999999673366767,
   0.9999999779933862,0.9999999854247601,
   0.9999999905209599,0.9999999939539325,
   0.9999999962228177,0.9999999976920703,
   0.9999999986229456,0.999999999199058,
   0.9999999995467416,0.9999999997509569,
   0.9999999998674497,0.9999999999318365,
   0.9999999999662286,0.9999999999839305,
   0.9999999999926823,0.9999999999968237,
   0.9999999999986915,0.9999999999994911,
   0.999999999999814,0.9999999999999367,
   0.99999999999998,0.9999999999999943,
   0.9999999999999983,0.9999999999999997,
   0.9999999999999998,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,
   1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.};
  
  for(i=1;i<=999;i++) {

    if (fabs(new_ascent_probability(1000,i) - expected_values1000[i-1]) > 1e-12) {
      
      printf("FAIL\n\t");
      printf("position %d (%g) doesn't match expected (%g)\n",i,
	     new_ascent_probability(1000,i),expected_values1000[i-1]);
      return false;
    }
    
  }

  printf("pass\n");

  printf("----------------------------------------------\n"
	 "testing new_ascent_probability values: pass  \n"
	 "----------------------------------------------\n");

  return true;
 
}


bool build_choice_vector_test() {

  printf("---------------------------------------------\n"
	 "testing build choice vector code             \n"
	 "---------------------------------------------\n");

  clock_t start,end;
  double cpu_time_used;

  const gsl_rng_type * rng_T;
  gsl_rng *rng;
  
  gsl_rng_env_setup();  
  rng_T = gsl_rng_default;
  rng = gsl_rng_alloc(rng_T);
  
  int seedi;
  seedi = time(0); 
  gsl_rng_set(rng,seedi);

  printf("testing generation of random choice vectors\n");   
  printf("with %s random number gen, seeded with %d.\n",
	 gsl_rng_name(rng),seedi);

  printf("generating some 10 choice vectors for inspection...\n");

  int i;
  for(i=0;i<10;i++) {

    perm_choice_t *cv;
    int nchoices;
    
    build_choice_vector(10,5,
			&nchoices,&cv,
			rng);

    int j;

    if (nchoices > 10 || nchoices < 5) {

      printf("FAIL (length of cv = %d > 10 or < 5)\n",
	     nchoices);
      return false;

    }

    for(j=nchoices-1;j>=0;j--) {

      printf(" {%d,%d}%s",cv[j].n,cv[j].k,
	     j > 0 ? "," : "\n");

    }

    if (cv[nchoices-1].k != cv[nchoices-1].n - 1 &&
	cv[nchoices-1].k != 0) {

      printf("FAIL (choice vector ends with {%d,%d} != {n,n-1} or {n,0}) \n",
	     cv[nchoices-1].n,cv[nchoices-1].k);
      return false;
      
    }

    if (cv[0].n != 10 || cv[0].k != 5) {

      printf("FAIL (choice vector starts with {%d,%d} != {10,5})\n",
	     cv[0].n,cv[0].k);
      return false;

    }

    free(cv);

  }

  printf("generating (4,2) choice vecs...");
  int starts_with_ascending = 0;
  int starts_with_descending = 0;
  
  for(i=0;i<10000;i++) {

    perm_choice_t *cv;
    int nchoices;
    
    build_choice_vector(4,2,
			&nchoices,&cv,
			rng);

    if (nchoices > 4 || nchoices < 2) {

      printf("FAIL (length of cv = %d > 4 or < 2)\n",
	     nchoices);
      return false;

    }
    
    if (cv[nchoices-1].k != cv[nchoices-1].n - 1 &&
	cv[nchoices-1].k != 0) {

      printf("FAIL (choice vector ends with {%d,%d} != {n,n-1} or {n,0}) \n",
	     cv[nchoices-1].n,cv[nchoices-1].k);
      return false;
      
    }

    if (cv[0].n != 4 || cv[0].k != 2) {

      printf("FAIL (choice vector starts with {%d,%d} != {4,2})\n",
	     cv[0].n,cv[0].k);
      return false;

    }

    if (cv[nchoices-1].k == cv[nchoices-1].n - 1) {

      starts_with_ascending++;

    }

    if (cv[nchoices-1].k == 0) {

      starts_with_descending++;
      
    }

    free(cv);

  }

  printf("(%d start with ascending, %d with descending)",
	 starts_with_ascending,starts_with_descending);

  if (starts_with_ascending == 0 || starts_with_descending == 0) {

    printf("FAIL (some of each should be generated)\n");
    return false;

  }

  printf("pass\n");

  printf("performance testing\n");
  printf("got 0.024 sec in Mathematica (2014) for (1000,500) vectors...");

  start = clock();

  for(i=0;i<1000;i++) {

    perm_choice_t *cv;
    int nchoices;
    
    build_choice_vector(1000,500,
			&nchoices,&cv,
			rng);

    free(cv);

  }

  end = clock();
  cpu_time_used = (((double)(end - start))/CLOCKS_PER_SEC)/1000.0;

  printf("%g sec (for this system)\n",cpu_time_used);    
  gsl_rng_free(rng);

  printf("---------------------------------------------\n"
	 "testing build choice vector code: pass       \n"
	 "---------------------------------------------\n");

  return true;

}

typedef struct permnode_struct {

  int s;
  int next;
  
} perm_node_t;


typedef struct workperm_struct {

  perm_node_t *node;
  int start;
  int max;
  int used;

} working_perm_t;

working_perm_t wperm_new(int n);
void make_descending_wperm(working_perm_t *wp,int n);
void make_ascending_wperm(working_perm_t *wp,int n);
void convert_wperm_to_buffer(working_perm_t *wp,int *buf);
void wperm_free(working_perm_t *wp);

void wperm_insert_at_ascent(working_perm_t *wp,
			    int toinsert,
			    int ascent);

void wperm_insert_at_descent(working_perm_t *wp,
			     int toinsert,
			     int descent);

void wperm_insert_at_start(working_perm_t *wp,
			   int toinsert);

void wperm_insert_at_end(working_perm_t *wp,
			 int toinsert);

int  ascent_count(int *perm,int n);

int int_cmp(const void *A,const void *B)
{
  int *a = (int *)(A);
  int *b = (int *)(B);

  return *a - *b;
}

bool isperm(int *buf,int n)
/* Tests whether buf is a permutation of 1..n */
{
  int *newbuf;
  newbuf = malloc(n*sizeof(int));
  assert(newbuf != NULL);

  int i;
  for(i=0;i<n;i++) {

    newbuf[i] = buf[i];

  }

  qsort(newbuf,n,sizeof(int),int_cmp);

  for(i=0;i<n;i++) {

    if (newbuf[i] != i+1) {

      free(newbuf);
      return false;

    }

  }

  free(newbuf);
  return true;
}

bool test_workperm() {

  printf("----------------------------------------\n"
	 "testing working perm linked list code   \n"
	 "----------------------------------------\n");

  printf("generating new 100 element wp...");
  working_perm_t wp = wperm_new(100);
  printf("pass (didn't crash)\n");

  printf("generating 22 element descending perm...");
  make_descending_wperm(&wp,22);

  int buf[100];
  convert_wperm_to_buffer(&wp,buf);

  int i;
  for(i=0;i<22;i++) {

    if (buf[i] != 22-i) {

      printf("FAIL (element %d == %d != %d)\n",
	     i,buf[i],22-i);
      return false;

    }

  }

  if (wp.used != 22) {

    printf("FAIL (wrong # of elements %d used != 22)\n",wp.used);
    return false;

  }

  if (!isperm(buf,22)) {

    printf("FAIL (doesn't pass isperm)\n");
    return false;

  }

  if (ascent_count(buf,22) != 0) {

    printf("FAIL (should have 0 ascents, has %d)\n",
	   ascent_count(buf,22));
    return false;

  }
    
  printf("pass (matches expected permutation)\n");
  
  printf("freeing wp...");
  wperm_free(&wp);
  wperm_free(&wp);
  printf("pass (passes double-free test, didn't crash)\n");

  printf("generating new 100 element wp...");
  wp = wperm_new(100);
  printf("pass (didn't crash)\n");

  printf("generating 43 element ascending perm...");
  make_ascending_wperm(&wp,43);

  convert_wperm_to_buffer(&wp,buf);

  for(i=0;i<43;i++) {

    if (buf[i] != i+1) {

      printf("FAIL (element %d == %d != %d)\n",
	     i,buf[i],i+1);
      return false;

    }

  }

 
  if (wp.used != 43) {

    printf("FAIL (wrong # of elements %d used != 43)\n",wp.used);
    return false;

  }

  if (!isperm(buf,43)) {

    printf("FAIL (doesn't pass isperm)\n");
    return false;

  }

  if (ascent_count(buf,43) != 42) {

    printf("FAIL (should have 42 ascents, has %d)\n",
	   ascent_count(buf,43));
    return false;

  }
  
  printf("pass (matches expected permutation)\n");
  
  printf("freeing wp...");
  wperm_free(&wp);
  wperm_free(&wp);
  printf("pass (passes double-free test, didn't crash)\n");

 
  printf("inserting 99 by ascent positions in 43 element ascent...");

  for(i=0;i<42;i++) {
    
    wp = wperm_new(99);
    make_ascending_wperm(&wp,43);
    wperm_insert_at_ascent(&wp,99,i);

    int buf[100];
    convert_wperm_to_buffer(&wp,buf);

    int j;
    for(j=0;j<44;j++) {

      if (j<=i) { /* We haven't inserted yet */

	if (buf[j] != j+1) {

	  printf("FAIL (inserted at ascent %d but buf[%d] = %d != %d)\n",
		 i,j,buf[j],j);
	  return false;

	}

      } else if (j==i+1) {

	if (buf[j] != 99) {

	  printf("FAIL (inserted 99 at ascent %d but "
		 "buf[%d] = %d != 99)\n",
		 i,j,buf[j]);
	  return false;

	}

      } else { /* j > i + 1 */

	if (buf[j] != j) {

	  printf("FAIL (inserted at ascent %d but buf[%d] = %d != %d)\n",
		 i,j,buf[j],j-1);
	  return false;

	}

      }

    }

    if (ascent_count(buf,44) != 42) {
      /* We inserted at an ascent, so shouldn't have generated one. */

      printf("FAIL (should have 42 ascents, but have %d)\n",
	     ascent_count(buf,44));
      return false;

    }

    wperm_free(&wp);

  }

  printf("pass (inserted 99 at positions 0..41 correctly)\n");

  printf("inserting 99 by descent positions in 43 element descent...");

  for(i=0;i<42;i++) {
    
    wp = wperm_new(99);
    make_descending_wperm(&wp,43);
    wperm_insert_at_descent(&wp,99,i);

    int buf[100];
    convert_wperm_to_buffer(&wp,buf);

    int j;
    for(j=0;j<44;j++) {

      if (j<=i) { /* We haven't inserted yet */

	if (buf[j] != 43-j) {

	  printf("FAIL (inserted at ascent %d but buf[%d] = %d != %d)\n",
		 i,j,buf[j],43-j);
	  return false;

	}

      } else if (j==i+1) {

	if (buf[j] != 99) {

	  printf("FAIL (inserted 99 at ascent %d but "
		 "buf[%d] = %d != 99)\n",
		 i,j,buf[j]);
	  return false;

	}

      } else { /* j > i + 1 */

	if (buf[j] != 43-j+1) {

	  printf("FAIL (inserted at ascent %d but buf[%d] = %d != %d)\n",
		 i,j,buf[j],43-j-1);
	  return false;

	}

      }

    }

    if (ascent_count(buf,44) != 1) {
      /* We inserted at a descent, so should have generated one. */

      printf("FAIL (should have 1 ascents, but have %d)\n",
	     ascent_count(buf,44));
      return false;

    }

    wperm_free(&wp);

  }

  printf("pass (inserted 99 at positions 0..41 correctly)\n");

  printf("testing insert of 99 at start in 43 element perm...");

  {
    working_perm_t wp;
    
    wp = wperm_new(99);
    make_ascending_wperm(&wp,43);
    wperm_insert_at_start(&wp,99);

    int buf[100];
    convert_wperm_to_buffer(&wp,buf);

    int j;
    for(j=0;j<44;j++) {

      if (j==0) {

	if (buf[j] != 99) {

	  printf("FAIL (inserted 99 at start but buf[0] = %d != 99)\n",
		 buf[0]);
	  return false;

	}

      } else {

	if (buf[j] != j) {

	  printf("FAIL (inserted 99 at start but "
		 "buf[%d] = %d != %d)\n",
		 j,buf[j],j+1);
	  return false;

	}

      }

    }

    if (ascent_count(buf,44) != 42) {
      /* We inserted at start, so shouldn't have generated one. */

      printf("FAIL (should have 42 ascents, but have %d)\n",
	     ascent_count(buf,44));
      return false;

    }


    wperm_free(&wp);

  }

  printf("pass\n");

  printf("testing insert of 99 at end in 43 element perm...");

  {
    working_perm_t wp;
    
    wp = wperm_new(99);
    make_ascending_wperm(&wp,43);
    wperm_insert_at_end(&wp,99);

    int buf[100];
    convert_wperm_to_buffer(&wp,buf);

    int j;
    for(j=0;j<44;j++) {

      if (j==43) {

	if (buf[j] != 99) {

	  printf("FAIL (inserted 99 at end but buf[43] = %d != 99)\n",
		 buf[43]);
	  return false;

	}

      } else {

	if (buf[j] != j+1) {

	  printf("FAIL (inserted 99 at end but "
		 "buf[%d] = %d != %d)\n",
		 j,buf[j],j+1);
	  return false;

	}

      }

    }

    if (ascent_count(buf,44) != 43) {
      /* We inserted at end, so should have generated one. */

      printf("FAIL (should have 43 ascents, but have %d)\n",
	     ascent_count(buf,44));
      return false;

    }

    wperm_free(&wp);

  }

  printf("pass\n");

  printf("----------------------------------------\n"
	 "working perm linked list code: pass     \n"
	 "----------------------------------------\n");

  return true;
}

typedef struct p6_struct {
  int e[6];
} perm6;

static perm6 all63perms[302] =
  {{{1, 2, 3, 6, 5, 4}},
   {{1, 2, 4, 3, 6, 5}},
   {{1, 2, 4, 6, 5, 3}},
   {{1, 2, 5, 3, 6, 4}},
   {{1, 2, 5, 4, 3, 6}},
   {{1, 2, 5, 4, 6, 3}},
   {{1, 2, 5, 6, 4, 3}},
   {{1, 2, 6, 3, 5, 4}},
   {{1, 2, 6, 4, 3, 5}},
   {{1, 2, 6, 4, 5, 3}},
   {{1, 2, 6, 5, 3, 4}},
   {{1, 3, 2, 4, 6, 5}},
   {{1, 3, 2, 5, 4, 6}},
   {{1, 3, 2, 5, 6, 4}},
   {{1, 3, 2, 6, 4, 5}},
   {{1, 3, 4, 2, 6, 5}},
   {{1, 3, 4, 6, 5, 2}},
   {{1, 3, 5, 2, 6, 4}},
   {{1, 3, 5, 4, 2, 6}},
   {{1, 3, 5, 4, 6, 2}},
   {{1, 3, 5, 6, 4, 2}},
   {{1, 3, 6, 2, 5, 4}},
   {{1, 3, 6, 4, 2, 5}},
   {{1, 3, 6, 4, 5, 2}},
   {{1, 3, 6, 5, 2, 4}},
   {{1, 4, 2, 3, 6, 5}},
   {{1, 4, 2, 5, 3, 6}},
   {{1, 4, 2, 5, 6, 3}},
   {{1, 4, 2, 6, 3, 5}},
   {{1, 4, 3, 2, 5, 6}},
   {{1, 4, 3, 5, 2, 6}},
   {{1, 4, 3, 5, 6, 2}},
   {{1, 4, 3, 6, 2, 5}},
   {{1, 4, 5, 2, 6, 3}},
   {{1, 4, 5, 3, 2, 6}},
   {{1, 4, 5, 3, 6, 2}},
   {{1, 4, 5, 6, 3, 2}},
   {{1, 4, 6, 2, 5, 3}},
   {{1, 4, 6, 3, 2, 5}},
   {{1, 4, 6, 3, 5, 2}},
   {{1, 4, 6, 5, 2, 3}},
   {{1, 5, 2, 3, 6, 4}},
   {{1, 5, 2, 4, 3, 6}},
   {{1, 5, 2, 4, 6, 3}},
   {{1, 5, 2, 6, 3, 4}},
   {{1, 5, 3, 2, 4, 6}},
   {{1, 5, 3, 4, 2, 6}},
   {{1, 5, 3, 4, 6, 2}},
   {{1, 5, 3, 6, 2, 4}},
   {{1, 5, 4, 2, 3, 6}},
   {{1, 5, 4, 6, 2, 3}},
   {{1, 5, 6, 2, 4, 3}},
   {{1, 5, 6, 3, 2, 4}},
   {{1, 5, 6, 3, 4, 2}},
   {{1, 5, 6, 4, 2, 3}},
   {{1, 6, 2, 3, 5, 4}},
   {{1, 6, 2, 4, 3, 5}},
   {{1, 6, 2, 4, 5, 3}},
   {{1, 6, 2, 5, 3, 4}},
   {{1, 6, 3, 2, 4, 5}},
   {{1, 6, 3, 4, 2, 5}},
   {{1, 6, 3, 4, 5, 2}},
   {{1, 6, 3, 5, 2, 4}},
   {{1, 6, 4, 2, 3, 5}},
   {{1, 6, 4, 5, 2, 3}},
   {{1, 6, 5, 2, 3, 4}},
   {{2, 1, 3, 4, 6, 5}},
   {{2, 1, 3, 5, 4, 6}},
   {{2, 1, 3, 5, 6, 4}},
   {{2, 1, 3, 6, 4, 5}},
   {{2, 1, 4, 3, 5, 6}},
   {{2, 1, 4, 5, 3, 6}},
   {{2, 1, 4, 5, 6, 3}},
   {{2, 1, 4, 6, 3, 5}},
   {{2, 1, 5, 3, 4, 6}},
   {{2, 1, 5, 6, 3, 4}},
   {{2, 1, 6, 3, 4, 5}},
   {{2, 3, 1, 4, 6, 5}},
   {{2, 3, 1, 5, 4, 6}},
   {{2, 3, 1, 5, 6, 4}},
   {{2, 3, 1, 6, 4, 5}},
   {{2, 3, 4, 1, 6, 5}},
   {{2, 3, 4, 6, 5, 1}},
   {{2, 3, 5, 1, 6, 4}},
   {{2, 3, 5, 4, 1, 6}},
   {{2, 3, 5, 4, 6, 1}},
   {{2, 3, 5, 6, 4, 1}},
   {{2, 3, 6, 1, 5, 4}},
   {{2, 3, 6, 4, 1, 5}},
   {{2, 3, 6, 4, 5, 1}},
   {{2, 3, 6, 5, 1, 4}},
   {{2, 4, 1, 3, 6, 5}},
   {{2, 4, 1, 5, 3, 6}},
   {{2, 4, 1, 5, 6, 3}},
   {{2, 4, 1, 6, 3, 5}},
   {{2, 4, 3, 1, 5, 6}},
   {{2, 4, 3, 5, 1, 6}},
   {{2, 4, 3, 5, 6, 1}},
   {{2, 4, 3, 6, 1, 5}},
   {{2, 4, 5, 1, 6, 3}},
   {{2, 4, 5, 3, 1, 6}},
   {{2, 4, 5, 3, 6, 1}},
   {{2, 4, 5, 6, 3, 1}},
   {{2, 4, 6, 1, 5, 3}},
   {{2, 4, 6, 3, 1, 5}},
   {{2, 4, 6, 3, 5, 1}},
   {{2, 4, 6, 5, 1, 3}},
   {{2, 5, 1, 3, 6, 4}},
   {{2, 5, 1, 4, 3, 6}},
   {{2, 5, 1, 4, 6, 3}},
   {{2, 5, 1, 6, 3, 4}},
   {{2, 5, 3, 1, 4, 6}},
   {{2, 5, 3, 4, 1, 6}},
   {{2, 5, 3, 4, 6, 1}},
   {{2, 5, 3, 6, 1, 4}},
   {{2, 5, 4, 1, 3, 6}},
   {{2, 5, 4, 6, 1, 3}},
   {{2, 5, 6, 1, 4, 3}},
   {{2, 5, 6, 3, 1, 4}},
   {{2, 5, 6, 3, 4, 1}},
   {{2, 5, 6, 4, 1, 3}},
   {{2, 6, 1, 3, 5, 4}},
   {{2, 6, 1, 4, 3, 5}},
   {{2, 6, 1, 4, 5, 3}},
   {{2, 6, 1, 5, 3, 4}},
   {{2, 6, 3, 1, 4, 5}},
   {{2, 6, 3, 4, 1, 5}},
   {{2, 6, 3, 4, 5, 1}},
   {{2, 6, 3, 5, 1, 4}},
   {{2, 6, 4, 1, 3, 5}},
   {{2, 6, 4, 5, 1, 3}},
   {{2, 6, 5, 1, 3, 4}},
   {{3, 1, 2, 4, 6, 5}},
   {{3, 1, 2, 5, 4, 6}},
   {{3, 1, 2, 5, 6, 4}},
   {{3, 1, 2, 6, 4, 5}},
   {{3, 1, 4, 2, 5, 6}},
   {{3, 1, 4, 5, 2, 6}},
   {{3, 1, 4, 5, 6, 2}},
   {{3, 1, 4, 6, 2, 5}},
   {{3, 1, 5, 2, 4, 6}},
   {{3, 1, 5, 6, 2, 4}},
   {{3, 1, 6, 2, 4, 5}},
   {{3, 2, 1, 4, 5, 6}},
   {{3, 2, 4, 1, 5, 6}},
   {{3, 2, 4, 5, 1, 6}},
   {{3, 2, 4, 5, 6, 1}},
   {{3, 2, 4, 6, 1, 5}},
   {{3, 2, 5, 1, 4, 6}},
   {{3, 2, 5, 6, 1, 4}},
   {{3, 2, 6, 1, 4, 5}},
   {{3, 4, 1, 2, 6, 5}},
   {{3, 4, 1, 5, 2, 6}},
   {{3, 4, 1, 5, 6, 2}},
   {{3, 4, 1, 6, 2, 5}},
   {{3, 4, 2, 1, 5, 6}},
   {{3, 4, 2, 5, 1, 6}},
   {{3, 4, 2, 5, 6, 1}},
   {{3, 4, 2, 6, 1, 5}},
   {{3, 4, 5, 1, 6, 2}},
   {{3, 4, 5, 2, 1, 6}},
   {{3, 4, 5, 2, 6, 1}},
   {{3, 4, 5, 6, 2, 1}},
   {{3, 4, 6, 1, 5, 2}},
   {{3, 4, 6, 2, 1, 5}},
   {{3, 4, 6, 2, 5, 1}},
   {{3, 4, 6, 5, 1, 2}},
   {{3, 5, 1, 2, 6, 4}},
   {{3, 5, 1, 4, 2, 6}},
   {{3, 5, 1, 4, 6, 2}},
   {{3, 5, 1, 6, 2, 4}},
   {{3, 5, 2, 1, 4, 6}},
   {{3, 5, 2, 4, 1, 6}},
   {{3, 5, 2, 4, 6, 1}},
   {{3, 5, 2, 6, 1, 4}},
   {{3, 5, 4, 1, 2, 6}},
   {{3, 5, 4, 6, 1, 2}},
   {{3, 5, 6, 1, 4, 2}},
   {{3, 5, 6, 2, 1, 4}},
   {{3, 5, 6, 2, 4, 1}},
   {{3, 5, 6, 4, 1, 2}},
   {{3, 6, 1, 2, 5, 4}},
   {{3, 6, 1, 4, 2, 5}},
   {{3, 6, 1, 4, 5, 2}},
   {{3, 6, 1, 5, 2, 4}},
   {{3, 6, 2, 1, 4, 5}},
   {{3, 6, 2, 4, 1, 5}},
   {{3, 6, 2, 4, 5, 1}},
   {{3, 6, 2, 5, 1, 4}},
   {{3, 6, 4, 1, 2, 5}},
   {{3, 6, 4, 5, 1, 2}},
   {{3, 6, 5, 1, 2, 4}},
   {{4, 1, 2, 3, 6, 5}},
   {{4, 1, 2, 5, 3, 6}},
   {{4, 1, 2, 5, 6, 3}},
   {{4, 1, 2, 6, 3, 5}},
   {{4, 1, 3, 2, 5, 6}},
   {{4, 1, 3, 5, 2, 6}},
   {{4, 1, 3, 5, 6, 2}},
   {{4, 1, 3, 6, 2, 5}},
   {{4, 1, 5, 2, 3, 6}},
   {{4, 1, 5, 6, 2, 3}},
   {{4, 1, 6, 2, 3, 5}},
   {{4, 2, 1, 3, 5, 6}},
   {{4, 2, 3, 1, 5, 6}},
   {{4, 2, 3, 5, 1, 6}},
   {{4, 2, 3, 5, 6, 1}},
   {{4, 2, 3, 6, 1, 5}},
   {{4, 2, 5, 1, 3, 6}},
   {{4, 2, 5, 6, 1, 3}},
   {{4, 2, 6, 1, 3, 5}},
   {{4, 3, 1, 2, 5, 6}},
   {{4, 3, 5, 1, 2, 6}},
   {{4, 3, 5, 6, 1, 2}},
   {{4, 3, 6, 1, 2, 5}},
   {{4, 5, 1, 2, 6, 3}},
   {{4, 5, 1, 3, 2, 6}},
   {{4, 5, 1, 3, 6, 2}},
   {{4, 5, 1, 6, 2, 3}},
   {{4, 5, 2, 1, 3, 6}},
   {{4, 5, 2, 3, 1, 6}},
   {{4, 5, 2, 3, 6, 1}},
   {{4, 5, 2, 6, 1, 3}},
   {{4, 5, 3, 1, 2, 6}},
   {{4, 5, 3, 6, 1, 2}},
   {{4, 5, 6, 1, 3, 2}},
   {{4, 5, 6, 2, 1, 3}},
   {{4, 5, 6, 2, 3, 1}},
   {{4, 5, 6, 3, 1, 2}},
   {{4, 6, 1, 2, 5, 3}},
   {{4, 6, 1, 3, 2, 5}},
   {{4, 6, 1, 3, 5, 2}},
   {{4, 6, 1, 5, 2, 3}},
   {{4, 6, 2, 1, 3, 5}},
   {{4, 6, 2, 3, 1, 5}},
   {{4, 6, 2, 3, 5, 1}},
   {{4, 6, 2, 5, 1, 3}},
   {{4, 6, 3, 1, 2, 5}},
   {{4, 6, 3, 5, 1, 2}},
   {{4, 6, 5, 1, 2, 3}},
   {{5, 1, 2, 3, 6, 4}},
   {{5, 1, 2, 4, 3, 6}},
   {{5, 1, 2, 4, 6, 3}},
   {{5, 1, 2, 6, 3, 4}},
   {{5, 1, 3, 2, 4, 6}},
   {{5, 1, 3, 4, 2, 6}},
   {{5, 1, 3, 4, 6, 2}},
   {{5, 1, 3, 6, 2, 4}},
   {{5, 1, 4, 2, 3, 6}},
   {{5, 1, 4, 6, 2, 3}},
   {{5, 1, 6, 2, 3, 4}},
   {{5, 2, 1, 3, 4, 6}},
   {{5, 2, 3, 1, 4, 6}},
   {{5, 2, 3, 4, 1, 6}},
   {{5, 2, 3, 4, 6, 1}},
   {{5, 2, 3, 6, 1, 4}},
   {{5, 2, 4, 1, 3, 6}},
   {{5, 2, 4, 6, 1, 3}},
   {{5, 2, 6, 1, 3, 4}},
   {{5, 3, 1, 2, 4, 6}},
   {{5, 3, 4, 1, 2, 6}},
   {{5, 3, 4, 6, 1, 2}},
   {{5, 3, 6, 1, 2, 4}},
   {{5, 4, 1, 2, 3, 6}},
   {{5, 4, 6, 1, 2, 3}},
   {{5, 6, 1, 2, 4, 3}},
   {{5, 6, 1, 3, 2, 4}},
   {{5, 6, 1, 3, 4, 2}},
   {{5, 6, 1, 4, 2, 3}},
   {{5, 6, 2, 1, 3, 4}},
   {{5, 6, 2, 3, 1, 4}},
   {{5, 6, 2, 3, 4, 1}},
   {{5, 6, 2, 4, 1, 3}},
   {{5, 6, 3, 1, 2, 4}},
   {{5, 6, 3, 4, 1, 2}},
   {{5, 6, 4, 1, 2, 3}},
   {{6, 1, 2, 3, 5, 4}},
   {{6, 1, 2, 4, 3, 5}},
   {{6, 1, 2, 4, 5, 3}},
   {{6, 1, 2, 5, 3, 4}},
   {{6, 1, 3, 2, 4, 5}},
   {{6, 1, 3, 4, 2, 5}},
   {{6, 1, 3, 4, 5, 2}},
   {{6, 1, 3, 5, 2, 4}},
   {{6, 1, 4, 2, 3, 5}},
   {{6, 1, 4, 5, 2, 3}},
   {{6, 1, 5, 2, 3, 4}},
   {{6, 2, 1, 3, 4, 5}},
   {{6, 2, 3, 1, 4, 5}},
   {{6, 2, 3, 4, 1, 5}},
   {{6, 2, 3, 4, 5, 1}},
   {{6, 2, 3, 5, 1, 4}},
   {{6, 2, 4, 1, 3, 5}},
   {{6, 2, 4, 5, 1, 3}},
   {{6, 2, 5, 1, 3, 4}},
   {{6, 3, 1, 2, 4, 5}},
   {{6, 3, 4, 1, 2, 5}},
   {{6, 3, 4, 5, 1, 2}},
   {{6, 3, 5, 1, 2, 4}},
   {{6, 4, 1, 2, 3, 5}},
   {{6, 4, 5, 1, 2, 3}},
   {{6, 5, 1, 2, 3, 4}}};

int perm63_cmp(const void *A,const void *B)
{
  perm6 *a = (perm6 *)(A);
  perm6 *b = (perm6 *)(B);

  int j;

  for(j=0;j<6;j++) {

    if (a->e[j] != b->e[j]) {

      return a->e[j] - b->e[j];

    }

  }

  return 0;
}

int which63perm(int *perm)
/* Identifies this permutation of 6 elements with 3 ascents using binary
   search on all63perms, using the fact that all63perms is sorted in 
   dictionary order. We'd do it brute-force, but we're going to call 
   this 302,000 times, at least. */
{
  perm6 *loc = bsearch(perm,all63perms,302,sizeof(perm6),perm63_cmp);
  if (loc != NULL) {

    int pos = loc - all63perms;
    assert(0 <= pos); /* By pointer arithmetic, should be position */
    assert(301 >= pos);
    return pos;

  }
  return -1; /* We didn't find it! */
}

typedef struct p4_struct {
  int e[4];
} perm4;

static perm4 all42perms[11] =
  {{{1, 2, 4, 3}},
   {{1, 3, 2, 4}},
   {{1, 3, 4, 2}},
   {{1, 4, 2, 3}},
   {{2, 1, 3, 4}},
   {{2, 3, 1, 4}},
   {{2, 3, 4, 1}},
   {{2, 4, 1, 3}},
   {{3, 1, 2, 4}},
   {{3, 4, 1, 2}},
   {{4, 1, 2, 3}}};

int perm42_cmp(const void *A,const void *B)
{
  perm4 *a = (perm4 *)(A);
  perm4 *b = (perm4 *)(B);

  int j;

  for(j=0;j<4;j++) {

    if (a->e[j] != b->e[j]) {

      return a->e[j] - b->e[j];

    }

  }

  return 0;
}

int which42perm(int *perm)
/* Identifies this permutation of 4 elements with 2 ascents using binary
   search on all42perms, using the fact that all42perms is sorted in 
   dictionary order. We'd do it brute-force, but we're going to call 
   this 11,000 times, at least. */
{
  perm4 *loc = bsearch(perm,all42perms,11,sizeof(perm4),perm42_cmp);
  if (loc != NULL) {

    int pos = loc - all42perms;
    assert(0 <= pos); /* By pointer arithmetic, should be position */
    assert(10 >= pos);
    return pos;

  }
  return -1; /* We didn't find it! */
}

bool test_k_ascent_permutation() {

  printf("---------------------------------------\n"
	 "random k ascent permutation tests      \n"
	 "---------------------------------------\n");
  
  clock_t start,end;
  double cpu_time_used;

  const gsl_rng_type * rng_T;
  gsl_rng *rng;
  
  gsl_rng_env_setup();  
  rng_T = gsl_rng_default;
  rng = gsl_rng_alloc(rng_T);
  
  int seedi;
  seedi = time(0); 
  gsl_rng_set(rng,seedi);

  printf("testing generation of random k ascent perms\n");   
  printf("with %s random number gen, seeded with %d.\n",
	 gsl_rng_name(rng),seedi);

  printf("generating some (10,5) permutations...\n");

  int i;
  for(i=0;i<10;i++) {

    int perm[10];
        
    random_k_ascent_permutation(10,5,perm,rng);

    int j;

    printf("\t");

    for(j=0;j<10;j++) {

      printf("%2d ",perm[j]);

    }

    printf("\n");

    if (!isperm(perm,10)) {

      printf("FAIL (doesn't pass isperm)\n");
      return false;

    }

    if (ascent_count(perm,10) != 5) {

      printf("FAIL (doesn't have 5 ascents, has %d)\n",
	     ascent_count(perm,10));
      return false;

    }

  }

  printf("testing database of 4 element permutations with 2 ascents...");
  for(i=0;i<11;i++) {

    if (!isperm(all42perms[i].e,4)) {

      printf("FAIL (permutation %d doesn't pass isperm)\n",i);
      return false;

    }

    if (ascent_count(all42perms[i].e,4) != 2) {

      printf("FAIL (permutation %d has %d != 2 ascents)\n",
	     i,ascent_count(all42perms[i].e,4));
      return false;

    }

  }
  printf("pass (all elts are permutations with 2 ascents)\n");

  printf("testing ability to locate perms in all42perms...");

  for(i=0;i<11;i++) {

    if (which42perm(all42perms[i].e) != i) {

      printf("FAIL (permutation %d appears in position %d)\n",
	     i,which42perm(all42perms[i].e));
      return false;

    }

  }
      
  printf("pass (was able to locate all perms)\n");

  printf("generating 11,000 permutations of 4 elts with 2 ascents...");
  int *tally;
  tally = calloc(11,sizeof(int));
  assert(tally != NULL);

  for(i=0;i<11000;i++) {

    int perm[4];
    random_k_ascent_permutation(4,2,perm,rng);

    if (!isperm(perm,4)) {

      printf("FAIL. Generated buffer ");
      int j;
      for(j=0;j<4;j++) { printf("%d ",perm[i]); }
      printf("which does not pass isperm\n");
      return false;

    }

    if (ascent_count(perm,4) != 2) {

      printf("FAIL. Generated buffer ");
      int j;
      for(j=0;j<4;j++) { printf("%d ",perm[i]); }
      printf("which has %d != 2 ascents\n",
	     ascent_count(perm,4));
      return false;

    }
     
    int pos;
    pos = which42perm(perm);

    if (pos == -1) {

      printf("FAIL. Generated 2 ascent permutation ");
      int j;
      for(j=0;j<4;j++) { printf("%d ",perm[i]); }
      printf(" which was not found in all42perms\n");
      return false;

    }

    tally[pos]++;

  }

  double normalized_tally[302];

  for(i=0;i<11;i++) {

    normalized_tally[i] = (double)(tally[i])/1000.0;

  }

  printf("\n\tnormalized tallies are\n\t   0: ");

  for(i=0;i<11;i++) {
    
    printf("%3.3f ",normalized_tally[i]);
    if((i+1)%10 == 0) { printf("\n\t %3d: ",i+1); }

  }

  printf("\n\t");
  
  int found_count=0;
  int missing_count=0;

  for(i=0;i<11;i++) {

    if (tally[i] == 0) { missing_count++; }
    else { found_count++; }

  }
	   
  printf("pass\n");

  printf("\n\t%d found permutations are\n\t\n",found_count);
  printf("\t{");

  int printed_count = 0;
  for(i=0;i<11;i++) {

    if (tally[i] != 0) {

      printf("%d",i+1);
      printed_count++;
      if (printed_count < found_count) { printf(","); }
      if ((printed_count+1) % 10 == 0) { printf("\n\t"); }

    }

  }
  printf("}\n");

  printf("\n\t%d missing permutations are\n\t\n",missing_count);
  printf("\t{");

  printed_count = 0;
  for(i=0;i<11;i++) {

    if (tally[i] == 0) {

      printf("%d",i+1);
      printed_count++;
      if (printed_count < missing_count) { printf(","); }
      if ((printed_count+1) % 10 == 0) { printf("\n\t"); }

    }

  }
  printf("}\n\n");
  free(tally);

  printf("testing database of 6 element permutations with 3 ascents...");
  for(i=0;i<302;i++) {

    if (!isperm(all63perms[i].e,6)) {

      printf("FAIL (permutation %d doesn't pass isperm)\n",i);
      return false;

    }

    if (ascent_count(all63perms[i].e,6) != 3) {

      printf("FAIL (permutation %d has %d != 3 ascents)\n",
	     i,ascent_count(all63perms[i].e,6));
      return false;

    }

  }
  printf("pass (all elts are permutations with 3 ascents)\n");

  printf("testing ability to locate perms in all63perms...");

  for(i=0;i<302;i++) {

    if (which63perm(all63perms[i].e) != i) {

      printf("FAIL (permutation %d appears in position %d)\n",
	     i,which63perm(all63perms[i].e));
      return false;

    }

  }
      
  printf("pass (was able to locate all perms)\n");

  printf("generating 302,000 permutations of 6 elts with 3 ascents...");
  tally = calloc(302,sizeof(int));
  assert(tally != NULL);

  for(i=0;i<302000;i++) {

    int perm[6];
    random_k_ascent_permutation(6,3,perm,rng);

    if (!isperm(perm,6)) {

      printf("FAIL. Generated buffer ");
      int j;
      for(j=0;j<6;j++) { printf("%d ",perm[i]); }
      printf("which does not pass isperm\n");
      return false;

    }

    if (ascent_count(perm,6) != 3) {

      printf("FAIL. Generated buffer ");
      int j;
      for(j=0;j<6;j++) { printf("%d ",perm[i]); }
      printf("which has %d != 3 ascents\n",
	     ascent_count(perm,6));
      return false;

    }
     
    int pos;
    pos = which63perm(perm);

    if (pos == -1) {

      printf("FAIL. Generated 3 ascent permutation ");
      int j;
      for(j=0;j<6;j++) { printf("%d ",perm[i]); }
      printf(" which was not found in all63perms\n");
      return false;

    }

    tally[pos]++;

  }

  for(i=0;i<302;i++) {

    normalized_tally[i] = (double)(tally[i])/1000.0;

  }

  printf("\n\tnormalized tallies are\n\t   0: ");

  for(i=0;i<302;i++) {
    
    printf("%3.3f ",normalized_tally[i]);
    if((i+1)%10 == 0) { printf("\n\t %3d: ",i+1); }

  }

  printf("\n\t");

  missing_count = 0;
  found_count = 0;

  for(i=0;i<302;i++) {

    if (tally[i] == 0) { missing_count++; }
    else { found_count++; }
    
  }

  printf("\n\t%d found permutations are\n\t\n",found_count);
  printf("\t{");

  printed_count = 0;
  for(i=0;i<302;i++) {

    if (tally[i] != 0) {

      printf("%3d",i+1);
      printed_count++;
      if (printed_count < found_count) { printf(","); }
      if ((printed_count+1) % 10 == 0) { printf("\n\t"); }

    }

  }
  printf("}\n");

  printf("\n\t%d missing permutations are\n\t\n",missing_count);
  printf("\t{");

  printed_count = 0;
  for(i=0;i<302;i++) {

    if (tally[i] == 0) {

      printf("%3d",i+1);
      printed_count++;
      if (printed_count < missing_count) { printf(","); }
      if ((printed_count+1) % 10 == 0) { printf("\n\t"); }

    }

  }
  printf("}\n\n");

  
  free(tally);
  printf("pass\n");

  printf("performance testing\n");
  printf("got 0.358 sec in Mathematica (2014) for (1000,500) permutations...");


  int perm[1000];
  start = clock();
  
  for(i=0;i<100;i++) {

    random_k_ascent_permutation(1000,500,perm,rng);

  }

  end = clock();
  cpu_time_used = (((double)(end - start))/CLOCKS_PER_SEC)/100.0;

  printf("%g sec (for this system)\n",cpu_time_used);    
  gsl_rng_free(rng);

  return true;

}
  

int main() {

  printf("test_k_ascent_permutations (%s)\n",PACKAGE_STRING);
  printf("--------------------------------------------------------\n"
	 "Unit tests for generating random k ascent permutations. \n"
	 "========================================================\n");

  if (!nap_lookup_table_test() || !build_choice_vector_test() ||
      !test_workperm() || !test_k_ascent_permutation()) {

    printf("=======================================================\n");
    printf("test_k_ascent_permutations:  FAIL.\n");
    exit(1);

  } else {

    printf("=====================================================\n");
    printf("test_k_ascent_permutations:  PASS.\n");
    exit(0);

  }

  return 0;

}
